# Before `make install' is performed this script should be runnable with
# `make test'. After `make install' it should work as `perl 1.t'
# $Id$
#########################

use strict;

use Test::More tests => 20;
BEGIN { use_ok('Class::SubclassDeep') };

my $init= init_top->new;
is( dump_object_simple( $init), "init_top: o1 => init_sub1 - o2 => init_sub2", "no mapping");

subclass( init_sub1 => "mapped_sub1", init_sub2 => "mapped_sub2");
my $mapped= init_top->new;
is( dump_object_simple( $mapped), "init_top: o1 => mapped_sub1 - o2 => mapped_sub2", "mapped 2 classes");

$init= init_top3->new();
is( dump_object_simple( $init), "init_top3: o3 => init_sub3", "no mapping, named constructor");
subclass( init_sub3 => "mapped_sub3 (constructor)");
$mapped= init_top3->new;
is( dump_object_simple( $mapped), "init_top3: o3 => mapped_sub3", "mapped, named constructor");

$init= init_top4->new;
is( dump_object( $init), 
    "init_top4: o1 => init_sub4_1 val_init_sub4_1 - o2 => init_sub4_2 val_init_sub4_2",
    "with vals, no mapping");

subclass( init_sub4_1 => "mapped_sub4_1", init_sub4_2 => "mapped_sub4_2");
$init= init_top4->new;
is( dump_object( $init), 
    "init_top4: o1 => mapped_sub4_1 val_mapped_sub4_1 - o2 => mapped_sub4_2 val_mapped_sub4_2",
    "with vals and constructors in mapped class");

$init= init_top5->new;
is( dump_object( $init), 
    "init_top5: o1 => init_sub5_1 val_init_sub5_1 - o2 => init_sub5_2 val_init_sub5_2",
    "no mapping");

subclass( init_sub5_1 => "mapped_sub5_1", init_sub5_2 => "mapped_sub5_2");
$mapped= init_top5->new;
is( dump_object( $mapped), 
    "init_top5: o1 => mapped_sub5_1 val_mapped_sub5_1 - o2 => mapped_sub5_2 val_mapped_sub5_2",
    "with vals and constructors using SUPER in mapped class");

$init= init_top6->new;
is( dump_object( $init), 
    "init_top6: o1 => init_sub6_1 new1 - o2 => init_sub6_1 new2",
    "2 constructors, no mapping");
subclass( init_sub6_1 => "mapped_sub6_1 (new1 new2)");
$mapped= init_top6->new;
is( dump_object( $mapped), 
    "init_top6: o1 => mapped_sub6_1 mnew1 - o2 => mapped_sub6_1 mnew2",
    "2 constructors, mapped, space separated");
subclass( init_sub6_1 => "mapped_sub6_1 (new1,new2)");
my $mapped2= init_top6->new;
is( dump_object( $mapped2), 
    "init_top6: o1 => mapped_sub6_1 mnew1 - o2 => mapped_sub6_1 mnew2",
    "2 constructors, mapped, comma separated");

$init= init_top7->new;
is( dump_object( $init), 
    "init_top7: o1 => init_sub7_1 val_1 - o2 => init_sub7_2 val_2",
    "subclasses inherit from base class, no mapping");
subclass( init_sub7_1 => "mapped_sub7_1", init_sub7_2 => "mapped_sub7_2");
$mapped= init_top7->new;
is( dump_object( $mapped), 
    "init_top7: o1 => mapped_sub7_1 mapped val_1 - o2 => mapped_sub7_2 mapped val_2",
    "subclasses inherit from base class, mapped");

$init= init_top8->new;
is( dump_object( $init), 
    "init_top8: o1 => init_sub8_1 val_1 - o2 => init_sub8_2 val_2",
    "subclasses inherit from base class, no mapping");
subclass( init_sub8_1 => "mapped_sub8_1", init_sub8_2 => "mapped_sub8_2");
$mapped= init_top8->new;
is( dump_object( $mapped), 
    "init_top8: o1 => mapped_sub8_1 val_3 - o2 => mapped_sub8_2 val_4",
    "mapped subclasses use sub classes, mapped");


$init= init_top9->new( "v1", "v2", "v3", "v4");
is( dump_object( $init),
    "init_top9: o1 => init_sub9_1 v1 v2 - o2 => init_sub9_2 v3 v4",
    "with arguments, no mapping");
subclass( init_sub9_1 => "mapped_sub9_1", init_sub9_2 => "mapped_sub9_2");
$mapped= init_top9->new( "v1", "v2", "v3", "v4");
is( dump_object( $mapped),
    "init_top9: o1 => mapped_sub9_1 mapped v1 v2 - o2 => mapped_sub9_2 mapped v3 v4",
    "with arguments, mapped");

$init= init_top10->new( "v1", "v2", "v3", "v4");
is( dump_object( $init),
    "init_top10: o1 => init_sub10_1 v1 v2 - o2 => init_sub10_2 v3 v4",
    "with arguments, no constructor, no mapping");
subclass( init_sub10_1 => "mapped_sub10_1", init_sub10_2 => "mapped_sub10_2");
$mapped= init_top10->new( "v5", "v6", "v7", "v8");
is( dump_object( $mapped),
    "init_top10: o1 => mapped_sub10_1 v5 v6 - o2 => mapped_sub10_2 v7 v8",
    "with arguments, no constructor mapped");



sub dump_object_simple # just the refs, not the content
  { my $o= shift;
    return ref($o) . ": " . join( " - ", map { "$_ => " . ref($o->{$_}) } sort keys %$o);
  }

sub dump_object       # refs AND content, to test constructors
  { my $o= shift;
    return ref($o) . ": " . join( " - ", map { "$_ => " . ref($o->{$_}) . " $o->{$_}->{val}" } sort keys %$o);
  }


package init_top;
sub new { return bless { o1 => init_sub1->new(), o2 => init_sub2->new }, shift(); }
package init_sub1;
sub new { return bless {}, shift(); }
package init_sub2;
sub new { return bless {}, shift(); }


package init_top3;
sub new { return bless { o3 => init_sub3->constructor() }, shift(); }
package init_sub3;
sub constructor { return bless {}, shift(); }

package init_top4;
sub new { return bless { o1 => init_sub4_1->new(), o2 => init_sub4_2->new }, shift(); }

package init_sub4_1;
sub new { return bless { val => "val_init_sub4_1" }, shift(); }
package init_sub4_2;
sub new { return bless {val => "val_init_sub4_2" }, shift(); }

package mapped_sub4_1;
use base 'init_sub4_1';
sub new { return bless { val => "val_mapped_sub4_1" }; }
package mapped_sub4_2;
use base 'init_sub4_2';
sub new { return bless { val => "val_mapped_sub4_2" }; }


package init_top5;
sub new { return bless { o1 => init_sub5_1->new(), o2 => init_sub5_2->new }, shift(); }
package init_sub5_1;
sub new { return bless { val => "val_init_sub5_1" }, shift(); }
package init_sub5_2;
sub new { return bless {val => "val_init_sub5_2" }, shift(); }

package mapped_sub5_1;
use base 'init_sub5_1';
sub new 
  { my $obj= mapped_sub5_1->SUPER::new( @_); 
    $obj->{val}= "val_mapped_sub5_1";
    return bless $obj;
  }
package mapped_sub5_2;
use base 'init_sub5_2';
sub new { my $obj= mapped_sub5_2->SUPER::new( @_);
          $obj->{val}= "val_mapped_sub5_2";
          return bless $obj;
        }


package init_top6;
sub new { return bless { o1 => init_sub6_1->new1(), o2 => init_sub6_1->new2 }, shift(); }
package init_sub6_1;
sub new1 { return bless { val => "new1"}; }
sub new2 { return bless { val => "new2"}; }
package mapped_sub6_1;
sub new1 { return bless { val => "mnew1"}; }
sub new2 { return bless { val => "mnew2"}; }


package init_top7;
sub new { return bless { o1 => init_sub7_1->new(), o2 => init_sub7_2->new }, shift(); }
package init_base7;
{ my $i; sub new { return bless { val => "val_" . ++$i }, shift(); } }
package init_sub7_1; use base 'init_base7';
package init_sub7_2; use base 'init_base7';

package mapped_base7;
{ my $i; sub new { return bless { val => "mapped val_" . ++$i }, shift() }; }
package mapped_sub7_1; use base 'mapped_base7';
package mapped_sub7_2; use base 'mapped_base7';


package init_top8;
sub new { return bless { o1 => init_sub8_1->new(), o2 => init_sub8_2->new }; }
package init_base8;
{ my $i; sub new { return bless { val => "val_" . ++$i }, shift() }; }
package init_sub8_1; use base 'init_base8';
package init_sub8_2; use base 'init_base8';

package mapped_sub8_1; use base 'init_sub8_1';
package mapped_sub8_2; use base 'init_sub8_2';


package init_top9;
sub new { my( $class, $v1, $v2, $v3, $v4)= @_; 
          return bless { o1 => init_sub9_1->new( $v1, $v2), o2 => init_sub9_2->new( $v3, $v4) }, $class;
        }

package init_sub9_1;
sub new { return bless { val => "$_[1] $_[2]" }, shift(); }
package init_sub9_2;
sub new { return bless { val => "$_[1] $_[2]"}, shift(); }


package mapped_sub9_1;
use base 'init_sub9_1';
sub new { return bless { val => "mapped $_[1] $_[2]" }; }
package mapped_sub9_2;
use base 'init_sub9_2';
sub new { return bless { val => "mapped $_[1] $_[2]" }; }


package init_top10;
sub new { my( $class, $v1, $v2, $v3, $v4)= @_; 
          return bless { o1 => init_sub10_1->new( $v1, $v2), o2 => init_sub10_2->new( $v3, $v4) }, $class;
        }

package init_sub10_1;
sub new { return bless { val => "$_[1] $_[2]" }, shift(); }
package init_sub10_2;
sub new { return bless { val => "$_[1] $_[2]"}, shift(); }

package mapped_sub10_1;
use base 'init_sub10_1';
package mapped_sub10_2;
use base 'init_sub10_2';

